package data;

import java.util.*;

import java.awt.*;

public class GridItem {

    public static final int TYPE_PLAIN  = 0;
    public static final int TYPE_WATER  = 1;
    public static final int TYPE_WALL   = 2;
    public static final int TYPE_HOME   = 3;

    public String theChar;
    public int type;
    public Hashtable robots = new Hashtable();
    public Hashtable packages = new Hashtable();
    public boolean explored = false;
    public int col = 0;
    public int row = 0;

    public String toString(boolean simple) {
        if (simple) {
            return theChar;
        }
        return ((robots.size() == 0) ? " " : "R") +
        theChar + ((packages.size() > 0) ? "*" : " ");
    }

    public Color getColor() {
        Color color = Color.lightGray;
        switch (type) {
        case TYPE_PLAIN:
            if (!explored) color = Color.gray;
            if (robots.size() == 1) {
                color = getRobot().color;
            }
            if (robots.size() > 1) color = Color.magenta;
            break;
        case TYPE_HOME:
            color = Color.yellow;
            if (robots.size() == 1) color = getRobot().color;
            if (robots.size() > 1) color = Color.magenta;
            break;
        case TYPE_WATER:
            color = Color.blue;
            break;
        case TYPE_WALL:
            color = Color.black;
            break;
        }
        return color;
    }
    
    public String key() {
        return col + "," + row;
    }
    
    public boolean isImpassible() {
        return ((type == TYPE_WATER) || (type == TYPE_WALL));
    }
    
    public GridItem(char item, int c, int r) {
        col = c;
        row = r;
        char[] chars = new char[1];
        chars[0] = item;
        theChar = new String(chars);
        switch (item) {
        case '.':
            type = TYPE_PLAIN;
            break;
        case '~':
            type = TYPE_WATER;
            explored = true;
            break;
        case '#':
            type = TYPE_WALL;
            explored = true;
            break;
        case '@':
            type = TYPE_HOME;
            break;
        }
    }

    public Enumeration packageElements() { return packages.elements(); }
                                      
    public PackageItem peekPackage(String id) {
        return (PackageItem) packages.get(id);
    }
    
    public PackageItem pickPackage(String id) {
        PackageItem item = (PackageItem) packages.get(id);
        packages.remove(id);
        return item;
    }
    
    public boolean dropPackage(PackageItem item) {
        if (item == null) return false;
        boolean delivered = false;
        if ((item.dCol == col) && (item.dRow == row)) {
            delivered = true;
        } else {
            packages.put(item.id, item);
        }
        return delivered;
    }
    
    public String getPackageString() {
        StringBuffer buf = new StringBuffer();
        for (Enumeration p = packages.elements(); p.hasMoreElements(); ) {
            PackageItem item = (PackageItem) p.nextElement();
            buf.append(item.id + " " + item.dCol + " " + item.dRow + " " + item.weight + " ");
        }
        return buf.toString();
    }
    
    public void addPackage(PackageItem p) {
        packages.put(p.id, p);
    }

    public boolean containsRobot(String id) {
        return ((robots.get(id) != null));
    }

    public Robot getRobot() {
        Iterator i = robots.values().iterator();
        Robot r = null;
        if (i.hasNext()) {
            r = (Robot) i.next();
        }
        return r;
    }
        
    public Robot getRobot(String id) {
        return (Robot) robots.get(id);
    }
        
    public boolean containsMyRobot(String id) {
        boolean me = false;
        Iterator i = robots.values().iterator();
        while (i.hasNext()) {
            Robot r = (Robot) i.next();
            if (r.me) {
                me = true;
                break;
            }
        }
        return me;
    }
    
    public boolean hasRobots() {
        return (robots.size() > 0);
    }
    
    public void addRobot(Robot r) {
        robots.put(r.robotID, r);
    }

    public void removeRobot(Robot r) {
        robots.remove(r.robotID);
    }
    
    public String listRobots() {
        Iterator i = robots.values().iterator();
        StringBuffer buf = new StringBuffer();
        while (i.hasNext()) {
            Robot r = (Robot) i.next();
            buf.append(r.toString() + " ");
        }
        return buf.toString();
    }
        
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append(theChar + " " + col + "," + row + " " + type + " " + listRobots());
        for (Enumeration p = packages.elements(); p.hasMoreElements();) {
            PackageItem item = (PackageItem) p.nextElement();
            buf.append(item.toString() + " ");
        }
        return buf.toString();
    }
}
