package data;

import java.util.*;

public class Path {

    public static Display display = null;
    
    private WorldMap worldMap;
    private int width;
    private int height;

    private Vector steps = new Vector();

    //    private Vector openSet = new Vector();
    private TreeSet   openSet;
    private Hashtable openHash = new Hashtable();
    private Hashtable closedSet = new Hashtable();

    private static final int[] offsetC = { -1, 1, 0,  0 };
    private static final int[] offsetR = {  0, 0, 1, -1 };

    private static Hashtable stepCache = new Hashtable();

    public Path(WorldMap map, int fCol, int fRow, int dCol, int dRow) {
        //System.out.println("PATH FROM " + fCol + "," + fRow + "  TO  " + dCol + "," + dRow);
        worldMap = map;
        width = worldMap.getWidth();
        height = worldMap.getHeight();

        String pathKey = fCol + "," + fRow + "  TO  " + dCol + "," + dRow;
        Vector cached = (Vector) stepCache.get(pathKey);
        if (cached != null) {
            steps = (Vector) cached.clone();
            //System.out.println("CACHED " + steps.size());
        } else {
            buildPath(fCol, fRow, dCol, dRow);
            Vector cache = (Vector) steps.clone();
            stepCache.put(pathKey, cache);
        }
    }

    protected void buildPath(int fCol, int fRow, int dCol, int dRow) {
        if (display != null) display.setDelivery(dCol, dRow);
        
        Step root = new Step(worldMap.getGridItem(fCol, fRow));
        openSet = new TreeSet(root);
        
        if (worldMap.getGridItem(dCol, dRow).isImpassible()) {
            System.out.println("CAN'T GO TO " + dCol + "," + dRow);
            return;
        }
        Step goal = null;
        openSet.add(root);
        openHash.put(stepKey(root), root);
        
        while (goal == null) {

            if (openSet.size() == 0) break;

            //debug("OPEN SET:", openSet);

            // BUILD FRONTIER SET
            while (openSet.size() > 0) {
                Step op = (Step) openSet.first();
                openSet.remove(op);

                if (closedSet.get(stepKey(op)) != null) continue;
                
                op.expanded = true;

                //System.out.println("EXPAND " + op);
                closedSet.put(stepKey(op), op);
                if ((display != null) && (op != root)) { display.updateGrid(op.col, op.row); }

                for (int c = 0; c < offsetC.length; c++) {
                    int nextC = op.col+offsetC[c];
                    int nextR = op.row+offsetR[c];

                    if ((nextC < 1) || (nextC > width) || (nextR < 1) || (nextR > height)) {
                        continue;
                    }

                    Step next = new Step(worldMap.getGridItem(nextC, nextR));
                    if (next.gridItem.isImpassible()) {
                        continue;
                    } else
                    if (closedSet.get(stepKey(next)) != null) {
                        continue;
                    } else {
                        float gcd = (float) dCol - next.col;
                        float grd = (float) dRow - next.row;
                        float enemy = 0.0f;
                        if ((next.gridItem != null) && next.gridItem.hasRobots()) {
                            enemy = 10.0f;
                        }
                        next.cost = op.cost + gcd*gcd + grd*grd + enemy;

                        next.parent = op;

                        String nextKey = stepKey(next);
                        Step previous = (Step) openHash.get(nextKey);
                        if (previous != null) {
                            if (next.cost < previous.cost) {
                                openSet.remove(previous);
                                openSet.add(next);
                            }
                        } else {
                            openHash.put(nextKey, next);
                            openSet.add(next);
                        }
                    }
                    
                    if ((next.col == dCol) && (next.row == dRow)) {
                        //System.out.println("GOAL!");
                        goal = next;
                        break;
                    }
                    if (goal != null) break;
                }
                if (goal != null) break;
            }
            if (goal != null) break;
        }

        if (goal != null) {
            Step step = goal;
            while (step.parent != null) {
                steps.insertElementAt(step, 0);
                if (display != null) display.setPath(step.col, step.row);
                step = step.parent;
            }
        }

        //System.out.println(this.toString());
        
        if (display != null) display.setPathLength(steps.size());
    }

    protected String stepKey(Step step) {
        return step.gridItem.col + "," + step.gridItem.row;
    }

    /*
    protected boolean insertInto(Step step, Vector set) {
        if (step.gridItem.isImpassible()) return false;
        
        boolean added = false;
        int from = 0;
        for (int i = 0; i < set.size(); i++) {
            Step existing = (Step) set.elementAt(i);
            if ((step.col == existing.col) && (step.row == existing.row)) {
                break;
            }
            if (step.cost < existing.cost) {
                from = i;
                set.insertElementAt(step, from);
                added = true;
                break;
            }
        }
        if (!added) {
            set.addElement(step);
        } else {
            for (int i = set.size() - 1; i > from; i--) {
                Step existing = (Step) set.elementAt(i);
                if ((step.col == existing.col) && (step.row == existing.row)) {
                    set.removeElementAt(i);
                    break;
                }
            }
        }

        return added;
    }
    */      

    public int length() { return steps.size(); }
    
    public GridItem lastStep() {
        Step step = null;
        if (steps.size() > 0) {
            step = (Step) steps.elementAt(steps.size()-1);
        }
        return step.gridItem;
    }
        
    public GridItem nextStep() {
        Step step = null;
        if (steps.size() > 0) {
            step = (Step) steps.elementAt(0);
            steps.removeElementAt(0);
            if (display != null) display.setPathLength(steps.size());
            return step.gridItem;
        }
        return null;
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("Steps: " + steps.size() + "\n");
        for (int i = 0; i < steps.size(); i++) {
            buf.append("  " + steps.elementAt(i) + "\n");
        }
        return buf.toString();
    }
}
